// calSupply.cc -*- c++ -*- Hans-Peter Stoerr Tue Feb 15 2000
// some additions to Cal-2.1

#include "calSupply.h"

#define HAVE_STDARG_H 1
#define HAVE_VARARGS_H 1
extern "C" {
#include "calInt.h"
}
#include <iostream>

// this is based on calBddSupport.c(c) 1994-1996 The Regents of the Univ. of California.

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void CSupBddVarCheckStep(Cal_BddManager_t * bddManager, Cal_Bdd_t f,
                                      CSupDependT *depends);
static void CSupBddUnmarkNodes(Cal_BddManager_t * bddManager, Cal_Bdd_t f);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

void
CSupBddVariableCheck(Cal_BddManager bddManager, Cal_Bdd fUserBdd,
               CSupVarCheckResT * varCheck)
{
  if(CalBddPreProcessing(bddManager, 1, fUserBdd)){
    Cal_Bdd_t f = CalBddGetInternalBdd(bddManager, fUserBdd);
    CSupDependT *depends = new CSupDependT[bddManager->numVars+1];
    int i;
    for (i=0;i<bddManager->numVars+1;i++) depends[i]=INDEP;
    CSupBddVarCheckStep(bddManager, f, depends);
    CSupBddUnmarkNodes(bddManager, f);
    for (i=0;i<bddManager->numVars+1;i++) if(INDEP!=depends[i]) {
      varCheck->bddVar = CalBddGetExternalBdd(bddManager, bddManager->varBdds[i]);
      varCheck->dep=depends[i];
      varCheck++;
    }
    delete depends;
  }
  varCheck->bddVar = (Cal_Bdd) 0;
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

static void
CSupBddVarCheckStep(
  Cal_BddManager_t * bddManager,
  Cal_Bdd_t  f,
  CSupDependT *depends)
{
  Cal_Bdd_t thenBdd,elseBdd;

  if(CalBddIsMarked(f) || CalBddIsBddConst(f)) // we have visited the node or it is trivial.
    return;

  CalBddMark(f);
  CalBddGetThenBdd(f, thenBdd);
  CalBddGetElseBdd(f, elseBdd);
  //now we should update depends[CalBddGetBddId(f)]
  assert(CalBddGetBddId(f)<=bddManager->numVars+1);
  if (BOTHDEP!=depends[CalBddGetBddId(f)]) { // else we there will be no change
    if (POSDEP!=depends[CalBddGetBddId(f)]) { // else this check was already run
      if (!CalBddIsBddZero(bddManager,thenBdd)) {
        if (INDEP==depends[CalBddGetBddId(f)]) depends[CalBddGetBddId(f)]=POSDEP;
        else depends[CalBddGetBddId(f)]=BOTHDEP;
      }
    } 
    if (NEGDEP!=depends[CalBddGetBddId(f)]) { // else this check was already run
      if (!CalBddIsBddZero(bddManager,elseBdd)) {
        if (INDEP==depends[CalBddGetBddId(f)]) depends[CalBddGetBddId(f)]=NEGDEP;
        else depends[CalBddGetBddId(f)]=BOTHDEP;
      }
    }
  }
  CSupBddVarCheckStep(bddManager, thenBdd, depends);
  CSupBddVarCheckStep(bddManager, elseBdd, depends);
}



/**Function********************************************************************

  Name        [CSupBddUnmarkNodes]

  Synopsis    [recursively unmarks the nodes]

  Description [optional]

  SideEffects [required]

  SeeAlso     [optional]

******************************************************************************/
static void
CSupBddUnmarkNodes(
  Cal_BddManager_t * bddManager,
  Cal_Bdd_t  f)
{
  Cal_Bdd_t tempBdd;

  if(!CalBddIsMarked(f) || CalBddIsBddConst(f)){
    return;
  }
  CalBddUnmark(f);
  tempBdd = bddManager->varBdds[CalBddGetBddId(f)];
  CalBddUnmark(tempBdd);
  CalBddGetThenBdd(f, tempBdd);
  CSupBddUnmarkNodes(bddManager, tempBdd);
  CalBddGetElseBdd(f, tempBdd);
  CSupBddUnmarkNodes(bddManager, tempBdd);
}


