// Hey Emacs! This is -*- c++ -*- ! :-)
// test.cc	-- Hans-Peter Stoerr Tue Dec 22 1998

#include <sys/times.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <deque>
#include <algo.h>
#include <multimap.h>
#include <math.h>
#include "objcal.h"

#if !defined(linux)
extern "C" { long random(); }
#endif

#define PR(X) do { (X).PrintInfo(#X); cout << flush; } while(0)
#define PR1(X) do { cout << #X << "=" << (X) << "\n" << flush; } while(0)

class Timer
{
public:
  Timer() { Restart(); }
  void Restart() { mStartTime=TheTime(); }
  operator double() { return TheTime()-mStartTime; }
private:                        
  double TheTime() { struct tms buf; times(&buf); return double(buf.tms_utime)/sysconf(_SC_CLK_TCK); }
  double mStartTime;
};

int main()
{
  Bdd b1("a"), b2("b"), b3("c"); // BDDs for variables a,b,c
  Bdd res = -(b1 * b2) + b3; // BDD for not(a /\ b) \/ c
  cout << res; // yields 

  const int numvar=25;
  const int numbdd=200;
  int i;
  for(i=0;i<numvar;i++) {
    char s[10];
    sprintf(s,"%d",i);
    mgr.Var(s);
  }
  BddVector vec;
  for(i=0;i<numbdd;++i)
    vec.push_back(mgr.Random(0.001));

  // Bdd theres=MultiwayOr(vec);
  {
  Timer tim;
  Bdd res;

  // res=mgr.Zero();
  // for(i=0;i<numbdd;i++) res+=vec[i];

//   deque<Bdd> queue;
//   copy(vec.begin(),vec.end(),back_inserter(queue));
//   while (queue.size()>=2) {
//     Bdd a(queue.front()); queue.pop_front();
//     Bdd b(queue.front()); queue.pop_front();
//     Bdd ab=a+b;
//     queue.push_back(ab);
//   }
//   res=queue.back();

  //   res=MultiwayOr(vec);

//   multimap<int,Bdd>   memory;
//   BddVector::const_iterator it=vec.begin(),end=vec.end();
//   for (;end!=it; ++it) {
//     memory.insert(make_pair(it->Size(),*it));
//   }
//   while (1<memory.size()) {
//     multimap<int,Bdd>::iterator it;
//     it=memory.begin();
//     Bdd first=it->second; memory.erase(it);
//     it=memory.begin();
//     Bdd second=it->second; memory.erase(it);
//     Bdd both=first+second;
//     memory.insert(make_pair(both.Size(),both));
//   }
//   res=memory.begin()->second;
  
  cout << tim << "\n";
  // assert(res==theres);
  PR(res);
  }
  return 0;
}

void    SuperCubeTest()
{
  Bdd   a("a"),b("b"),c("c"),d("d"),e("e"),f("f"),g("g"),h("h");
  Bdd   t=a*-b*-d+ -a*-f*(g+-h);
  BddVector     v;
  cout << t.SuperCube() << "\n";
  cout << t << "\n";
  v=t.Support();
  cout << v.size() << "\n";
}

string  MakeString(int i)
{
  char  s[1000];
  sprintf(s,"v%02d",i);
  return s;
}

void    OverflowTest()
{
  int   varnum,i;
  Bdd   riesiggross=mgr.Zero();
  for(varnum=1;varnum<20;varnum++)
    Bdd newvar(MakeString(varnum));
  for(;1;varnum++) {
    Bdd newvar(MakeString(varnum));
    for (i=0;i<4;i++) {
      Bdd rand_bdd;
      rand_bdd=mgr.XorRandom(varnum*varnum/3+1);
      rand_bdd.PrintInfo();
      cout << "( " << flush; 
      long olim=mgr.SetNodeLimit(10);
      Bdd warnix = riesiggross^rand_bdd;
      mgr.SetNodeLimit(olim); mgr.WasOverflow(); cout << ")\n" << flush;
      riesiggross ^= rand_bdd;
      if (mgr.WasOverflow()) return;
      riesiggross.PrintInfo();
    }
  }
}

static bool rndBool()
{
  long res=random();
  return ((res*(res/1<<16))>>16)%2;
}

void RunOverflowtest() 
{
  mgr.SetNodeLimit(100000);
  OverflowTest();
  mgr.SetNodeLimit(10);
  mgr.SetNodeLimit(300000);
  cout << "--------------------------\n";
  OverflowTest();
}

Bdd     RandCube(int n,BddVector& support)
{                               // returns random cube of size n.
  Bdd res=mgr.One();
  bool hadit[support.size()]; int i;
  for(i=0;i<int(support.size());i++) hadit[i]=0;
  for(i=0;i<n;i++) {
    int nvar;
    do {
      nvar=random()%support.size();
    } while( hadit[nvar] );
    if (rndBool()) res*=support[nvar];
    else res*=-support[nvar];
    ++hadit[nvar];
  }
  return res;
} 

void CalcRatios(double& maxRatio, double& avgRatio,
                int nVars,int nCubes,
                int nVarsCube,int nTries) 
{
  maxRatio=0; avgRatio=0;
  int i,j;
  for(i=0;i<nVars;i++) {Bdd x(MakeString(i));}
  BddVector support(mgr.NPrimedVars(0));
  for(i=0;i<nTries;i++) {
    Bdd test=mgr.Zero();
    int dnfsize=0;
    for(j=0;j<nCubes;j++) {
      Bdd ncube=RandCube(nVarsCube,support);
      dnfsize+=nVarsCube;
      test+=ncube; 
    }
    double ratio=double(test.Size())/dnfsize;
    avgRatio+=ratio;
    if(maxRatio<ratio) maxRatio=ratio;
  }
  avgRatio/=nTries;
}


int main1()
{
  double maxRatio,avgRatio,globmaxRatio=0;
  int nVars=40;
  int nCubes=10;
  int nVarsCube=8;
  int nTries=100;
  for(nCubes=2;nCubes<nVars*nVars/4;nCubes++){
    for(nVarsCube=2;nVarsCube<nVars-2;nVarsCube++) {
      CalcRatios(maxRatio,avgRatio,nVars,nCubes,nVarsCube,nTries);
      cout << maxRatio << "," << flush;
      if(globmaxRatio<maxRatio) globmaxRatio=maxRatio;
    }
    cout << "\n globmax: " << globmaxRatio << "\n";
  }
  return 0;
}

Bdd InterleaveCubes(int n,int m)
{ // interleave n cubes of m variables
  Bdd res=mgr.Zero();
  int i,j;
  for (i=0;i<n*m;i++) mgr.Var(MakeString(i));
  for (i=0;i<n;i++){
    Bdd cube=mgr.One();
    for (j=0;j<m;j++) cube*=mgr.Var(MakeString(i+j*m));
    res+=cube;
  }
  return res;
}

int InterleaveCubesmain()
{
  cout << InterleaveCubes(2,2) << "\n";
  for(int i=0;i<15;i++) {
    double siz=InterleaveCubes(i,i).Size();
    cout << i << ": " << siz << "\n";
    PR1(siz/i/i);
    PR1(log((double)siz)/log((double)i));
    PR1(log(siz)/i);
  }
  return 0;
}

int ExistsTimingsmain()
{
  int i,j;
  for (i=1;i<50;i++) Bdd x(MakeString(i));
  BddVector v10;
  for (i=20;i<30;i++) v10.push_back(Bdd(MakeString(i)));
  Bdd b=mgr.XorRandom(10);
  BddAssoc      ass(v10);
  Timer tim;
  for (j=0;j<1000;j++) {
    b.Exists(ass);
  }
  assert (b.Exists(ass)==b.Exists(v10));
  cout << tim << "\n";
  {
    Timer tim;
    for (j=0;j<1000;j++) {
      BddAssoc ass(v10);
      b.Exists(ass);
    }
    assert (b.Exists(ass)==b.Exists(v10));
    cout << tim << "\n";
  }
  {
    Timer tim;
    for (j=0;j<1000;j++) {
      b.Exists(v10);
    }
    assert (b.Exists(ass)==b.Exists(v10));
    cout << tim << "\n";
  }
  Timer tim1;
  for (j=0;j<1000;j++) {
    Bdd r=b;
    for (i=20;i<30;i++) { Bdd x(MakeString(i)); r=r.Exists(x); }
  }
  cout << tim1 << "\n";
  return 0;
}

int DeconstructBddsmain() 
{ // Schrott.
  int i,j=0;
  for (i=1;i<50;i++) Bdd x(MakeString(i));
  // srandom(time(NULL));
  Bdd test=mgr.XorRandom(20);
  while(test.Size()>2) {
    cout << test.Size();
    Bdd subtest=mgr.Zero();
    // for(int k=0;k<50;k++)
    // subtest+=test.RandomPath();
    cout << "\t" << subtest.Size();
    subtest=(subtest.Reduce(-test+subtest));
    cout << "->" << subtest.Size() << "\n";
    test=test*-subtest;
    j++;
  }
  cout << test << ":" << j << "\n";
  return 0;
}

void    CheckConstructions(const BddVector& v)
{
  int i,j;
  {
    Timer tim;
    Bdd res;
    for (i=0;i<100;i++)
      res=MultiwayOr(v);
    cout << tim << "\n";
  }
  {
    Timer tim;
    Bdd res;
    for (i=0;i<100;i++) {
      res=mgr.One();
      BddVector::const_iterator it=v.begin(),end=v.end();
      for (;end!=it;++it) res*= *it;
    }
    cout << tim << "\n";
  }
  {
    Timer tim;
    Bdd res;
    for (i=0;i<100;i++) {
      res=mgr.One();
      BddVector::const_iterator it=v.begin(),end=v.end();
      for (;end!=it;++it) res=ITE(*it,res,mgr.Zero());
    }
    cout << tim << "\n";
  }
  {
    Timer tim;
    Bdd res;
    for (i=0;i<100;i++) {
      deque<Bdd> queue;
      copy(v.begin(),v.end(),back_inserter(queue));
      while (queue.size()>=2) {
        Bdd a(queue.front()); queue.pop_front();
        Bdd b(queue.front()); queue.pop_front();
        Bdd ab=a+b;
        queue.push_back(ab);
      }
      res=queue.back();
    }
    cout << tim << "\n";
  }
}

struct ltstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

int nmain() 
{ // Schrott.
  const int maxvars=200;
  int i,j=0;
  BddVector     v;
  for (i=1;i<maxvars;i++) {
    Bdd x(MakeString(i));
    v.push_back(x);
  }

  CheckConstructions(v);
  reverse(v.begin(),v.end());
  CheckConstructions(v);
  random_shuffle(v.begin(),v.end());
  CheckConstructions(v);
  
  return 0;
}


